static void log_error(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " ERROR [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
  exit(EXIT_FAILURE);
}

static volatile int64_t limit_counter = 0;

static inline void log_limit(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  if(limit_counter++ % 1000000)
    return;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " DEBUG [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}

#if 1
static void log_info(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " INFO [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}
#else
static void log_info(const char * section, const char * msg, ...) {
  return;
}
#endif

#ifdef DEBUG
static void log_debug(const char * section, const char * msg, ...) {
  time_t t;
  struct tm ts;
  char line[256];
  va_list arglist;

  time(&t);
  ts = *localtime(&t);
  strftime(line, 80, "%m-%d %H:%M:%S", &ts);
  sprintf(line + strlen(line), " DEBUG [%s] ", section);
  va_start(arglist, msg);
  vsprintf(line + strlen(line), msg, arglist);
  va_end(arglist);
  puts(line);
}

#else
#define log_debug(...)
#endif

static int latency_target_fd = -1;
static int32_t latency_target_value = 0;

/* Latency trick
 * if the file /dev/cpu_dma_latency exists,
 * open it and write a zero into it. This will tell
 * the power management system not to transition to
 * a high cstate (in fact, the system acts like idle=poll)
 * When the fd to /dev/cpu_dma_latency is closed, the behavior
 * goes back to the system default.
 *
 * Documentation/power/pm_qos_interface.txt
 */
void set_latency_target(void) {
  struct stat s;
  int err;

  errno = 0;
  err = stat("/dev/cpu_dma_latency", &s);
  if (err == -1) {
    error(EXIT_FAILURE, errno, "WARN: stat /dev/cpu_dma_latency failed");
    return;
  }

  errno = 0;
  latency_target_fd = open("/dev/cpu_dma_latency", O_RDWR);
  if (latency_target_fd == -1) {
    error(EXIT_FAILURE, errno, "WARN: open /dev/cpu_dma_latency");
    return;
  }

  errno = 0;
  err = write(latency_target_fd, &latency_target_value, 4);
  if (err < 1) {
    error(EXIT_FAILURE, errno, "# error setting cpu_dma_latency to %d!",
        latency_target_value);
    close(latency_target_fd);
    return;
  }
  printf("# /dev/cpu_dma_latency set to %dus\n", latency_target_value);
}

// Timestamps utils
#define NSEC_PER_SEC INT64_C(1000000000)
static struct timespec int_to_ts(int64_t t) {
  struct timespec ts;
  ts.tv_sec = t / NSEC_PER_SEC;
  ts.tv_nsec = t - (ts.tv_sec * NSEC_PER_SEC);
  return ts;
}
static int64_t ts_to_int(struct timespec ts) {
  return ts.tv_sec * NSEC_PER_SEC + ts.tv_nsec;
}
static void add_ns(struct timespec *t, int64_t ns) {
  t->tv_nsec += ns;
  while (t->tv_nsec >= ((int64_t)NSEC_PER_SEC)) {
    t->tv_sec += 1;
    t->tv_nsec -= NSEC_PER_SEC;
  }
}
static int64_t calcdiff_ns(struct timespec t1, struct timespec t2) {
  int64_t diff;
  diff = NSEC_PER_SEC * ((int)t1.tv_sec - (int)t2.tv_sec);
  diff += ((int)t1.tv_nsec - (int)t2.tv_nsec);
  return diff;
}
